"""Synthesizes speech from the input string of text."""

from pathlib import Path
from typing import cast

import anyio
import click
import mutagen.mp3
from google.cloud import texttospeech

from .base import Speaker, common_options


class GoogleSpeaker(Speaker):
    """Google TTS speaker.

    Args:
        voice (str): The voice to use. Defaults to "en-US-Studio-M".
        speaking_rate (float):
            Optional. Input only. Speaking rate/speed, in the range
            [0.25, 4.0]. 1.0 is the normal native speed supported by the
            specific voice. 2.0 is twice as fast, and 0.5 is half as
            fast. If unset(0.0), defaults to the native 1.0 speed. Any
            other values < 0.25 or > 4.0 will return an error.
        pitch (float):
            Optional. Input only. Speaking pitch, in the range [-20.0,
            20.0]. 20 means increase 20 semitones from the original
            pitch. -20 means decrease 20 semitones from the original
            pitch.
        volume_gain_db (float):
            Optional. Input only. Volume gain (in dB) of the normal
            native volume supported by the specific voice, in the range
            [-96.0, 16.0]. If unset, or set to a value of 0.0 (dB), will
            play at normal native signal amplitude. A value of -6.0 (dB)
            will play at approximately half the amplitude of the normal
            native signal amplitude. A value of +6.0 (dB) will play at
            approximately twice the amplitude of the normal native
            signal amplitude. Strongly recommend not to exceed +10 (dB)
            as there's usually no effective increase in loudness for any
            value greater than that.
    """

    def __init__(
        self,
        *,
        voice: str = "en-US-Studio-M",
        speaking_rate: float = 1.0,
        pitch: float = 0.0,
        volume_gain_db: float = 0.0,
    ) -> None:
        self.speaking_rate = speaking_rate
        self.pitch = pitch
        self.voice = voice
        self.volume_gain_db = volume_gain_db

    async def synthesize(self, text: str, out_file: Path) -> float:
        input_text = texttospeech.SynthesisInput(text=text)
        audio_config = texttospeech.AudioConfig(
            audio_encoding=texttospeech.AudioEncoding.MP3,
            speaking_rate=self.speaking_rate,
            pitch=self.pitch,
            volume_gain_db=self.volume_gain_db,
        )
        voice = texttospeech.VoiceSelectionParams(
            language_code="-".join(self.voice.split("-")[:2]),
            name=self.voice,
        )
        async with texttospeech.TextToSpeechAsyncClient() as client:
            resp = await client.synthesize_speech(
                request={
                    "input": input_text,
                    "voice": voice,
                    "audio_config": audio_config,
                }
            )
        file = anyio.Path(out_file)
        async with await file.open("wb") as f:
            await f.write(resp.audio_content)

        audio = mutagen.mp3.MP3(out_file)
        return cast(float, audio.info.length)

    @classmethod
    def list_voices(cls) -> list[str]:
        with texttospeech.TextToSpeechClient() as client:
            resp = client.list_voices()
        return [v.name for v in resp.voices]

    @classmethod
    def get_command(cls) -> click.Command:
        @click.command()
        @click.option("--voice", default="en-US-Studio-M", help="The voice to use.")
        @click.option(
            "--speaking-rate", default=1.0, help="The speaking rate.", type=click.FLOAT
        )
        @click.option("--pitch", default=0.0, help="The pitch.", type=click.FLOAT)
        @click.option(
            "--volume-gain-db", default=0.0, help="The volume gain.", type=click.FLOAT
        )
        @common_options(cls)
        def google(
            voice: str,
            speaking_rate: float,
            pitch: float,
            volume_gain_db: float,
            text: str,
            output: str,
        ) -> None:
            speaker = cls(
                voice=voice,
                speaking_rate=speaking_rate,
                pitch=pitch,
                volume_gain_db=volume_gain_db,
            )
            speaker.say(text, Path(output))

        return google



from pathlib import Path
from typing import cast

import click
import mutagen.mp3

from .base import Speaker, common_options


class OpenAISpeaker(Speaker):
    """OpenAI TTS speaker.

    Args:
        model (str): The model to use. Defaults to "tts-1".
        voice (str): The voice to use. Defaults to "alloy".
        speed (float, optional): The speed of the speech.
        api_key (str, optional): The OpenAI API key.
        api_base (str, optional): The OpenAI API base URL.
    """

    def __init__(
        self,
        *,
        model: str = "tts-1",
        voice: str = "alloy",
        speed: float | None = None,
        api_key: str | None,
        api_base: str | None,
    ) -> None:
        import openai

        self.voice = voice
        self.speed = speed
        self.model = model
        self.client = openai.AsyncOpenAI(api_key=api_key, base_url=api_base)

    async def synthesize(self, text: str, out_file: Path) -> float:
        extra_args = {"speed": self.speed} if self.speed is not None else {}
        async with self.client.with_streaming_response.audio.speech.create(
            model=self.model,
            input=text,
            voice=self.voice,
            **extra_args,
        ) as resp:
            await resp.stream_to_file(out_file, chunk_size=8192)

        audio = mutagen.mp3.MP3(out_file)
        return cast(float, audio.info.length)

    @classmethod
    def list_voices(cls) -> list[str]:
        return ["alloy", "echo", "fable", "onyx", "nova", "shimmer"]

    @classmethod
    def get_command(cls) -> click.Command:
        @click.command()
        @click.option(
            "--api-key",
            required=True,
            envvar="OPENAI_API_KEY",
            help="The OpenAI API key.",
            show_envvar=True,
        )
        @click.option(
            "--api-base",
            envvar="OPENAI_API_BASE",
            help="The OpenAI API base URL.",
            show_envvar=True,
        )
        @click.option("--model", default="tts-1", help="The model to use.")
        @click.option("--speed", type=float, help="The speed of the speech.")
        @click.option("--voice", default="alloy", help="The voice to use.")
        @common_options(cls)
        def openai(
            api_key: str | None,
            api_base: str | None,
            model: str,
            speed: float | None,
            voice: str,
            text: str,
            output: str,
        ) -> None:
            speaker = cls(
                model=model,
                voice=voice,
                speed=speed,
                api_key=api_key,
                api_base=api_base,
            )
            speaker.say(text, Path(output))

        return openai



from dataclasses import dataclass
from pathlib import Path

import click

from .base import Speaker, SynthesizeError, common_options
from .consts import AZURE_SUPPORTED_VOICES


@dataclass
class EdgeSpeaker(Speaker):
    """Edge TTS speaker.

    Args:
        voice (str): The voice to use.
        rate (str): The rate of speech.
        pitch (str): The pitch of speech.
        volume (str): The volume of speech.
    """

    voice: str = "en-US-AriaNeural"
    rate: str = "+0%"
    pitch: str = "+0Hz"
    volume: str = "+0%"

    async def synthesize(self, text: str, out_file: Path) -> float:
        import edge_tts

        communicate = edge_tts.Communicate(
            text, voice=self.voice, rate=self.rate, pitch=self.pitch, volume=self.volume
        )
        duration = 0
        with open(out_file, "wb") as f:
            async for chunk in communicate.stream():
                if chunk["type"] == "audio":
                    f.write(chunk["data"])
                elif chunk["type"] == "WordBoundary":
                    duration = (chunk["offset"] + chunk["duration"]) / 1e7
            if duration == 0:
                raise SynthesizeError("Failed to get tts from edge")
        return duration

    @classmethod
    def list_voices(cls) -> list[str]:
        return AZURE_SUPPORTED_VOICES

    @classmethod
    def get_command(cls) -> click.Command:
        @click.command()
        @click.option("--voice", default="en-US-AriaNeural", help="The voice to use.")
        @click.option("--rate", default="+0%", help="The rate of speech.")
        @click.option("--pitch", default="+0Hz", help="The pitch of speech.")
        @click.option("--volume", default="+0%", help="The volume of speech.")
        @common_options(cls)
        def edge(
            voice: str, rate: str, pitch: str, volume: str, text: str, output: str
        ) -> None:
            speaker = cls(voice=voice, rate=rate, pitch=pitch, volume=volume)
            speaker.say(text, Path(output))

        return edge



import os
from pathlib import Path

import pytest


@pytest.mark.asyncio
async def test_openai_speaker(tmp_path: Path):
    from tetos.openai import OpenAISpeaker

    speaker = OpenAISpeaker(
        api_base=os.getenv("OPENAI_API_BASE"),
        api_key=os.getenv("OPENAI_API_KEY"),
    )
    duration = await speaker.synthesize("Hello, world!", tmp_path / "hello.mp3")
    assert 0.5 < duration < 4
    assert (tmp_path / "hello.mp3").stat().st_size > 0


@pytest.mark.asyncio
async def test_edge_speaker(tmp_path: Path):
    from tetos.edge import EdgeSpeaker

    speaker = EdgeSpeaker()
    duration = await speaker.synthesize("Hello, world!", tmp_path / "hello.mp3")
    assert 0.5 < duration < 4
    assert (tmp_path / "hello.mp3").stat().st_size > 0


@pytest.mark.asyncio
async def test_azure_speaker(tmp_path: Path):
    from tetos.azure import AzureSpeaker

    speaker = AzureSpeaker(
        speech_key=os.getenv("AZURE_SPEECH_KEY"),
        speech_region=os.getenv("AZURE_SPEECH_REGION"),
    )
    duration = await speaker.synthesize("Hello, world!", tmp_path / "hello.mp3")
    assert 0.5 < duration < 4
    assert (tmp_path / "hello.mp3").stat().st_size > 0


@pytest.mark.asyncio
async def test_volc_speaker(tmp_path: Path):
    from tetos.volc import VolcSpeaker

    speaker = VolcSpeaker(
        access_key=os.getenv("VOLC_ACCESS_KEY"),
        secret_key=os.getenv("VOLC_SECRET_KEY"),
        app_key=os.getenv("VOLC_APP_KEY"),
    )
    duration = await speaker.synthesize("Hello, world!", tmp_path / "hello.mp3")
    assert 0.5 < duration < 4
    assert (tmp_path / "hello.mp3").stat().st_size > 0



import os
from pathlib import Path

import anyio.to_thread
import azure.cognitiveservices.speech as speechsdk
import click

from .base import Speaker, SynthesizeError, common_options
from .consts import AZURE_SUPPORTED_VOICES


class AzureSpeaker(Speaker):
    """Azure TTS speaker.

    Args:
        speech_key (str): The Azure Speech key.
        speech_region (str): The Azure Speech region.
        voice (str): The voice to use.
    """

    def __init__(
        self, speech_key: str, speech_region: str, *, voice: str = "en-US-AriaNeural"
    ) -> None:
        self.speech_config = speechsdk.SpeechConfig(
            subscription=speech_key, region=speech_region
        )
        self.speech_config.speech_synthesis_voice_name = voice
        self.speech_config.set_speech_synthesis_output_format(
            speechsdk.SpeechSynthesisOutputFormat.Audio16Khz32KBitRateMonoMp3
        )
        self._set_proxy()

    def _set_proxy(self) -> None:
        from urllib.parse import urlparse

        for env in ("http_proxy", "https_proxy", "all_proxy"):
            # Try both lowercase and uppercase versions of the environment variable
            url = os.getenv(env, os.getenv(env.upper(), ""))
            if not url:
                continue
            parsed_url = urlparse(url)
            self.speech_config.set_proxy(
                parsed_url.hostname,
                parsed_url.port,
                parsed_url.username,
                parsed_url.password,
            )
            break

    async def synthesize(self, text: str, out_file: Path) -> float:
        audio_config = speechsdk.audio.AudioOutputConfig(filename=str(out_file))
        speech_synthesizer = speechsdk.SpeechSynthesizer(
            speech_config=self.speech_config, audio_config=audio_config
        )
        result = await anyio.to_thread.run_sync(speech_synthesizer.speak_text, text)

        if result.reason == speechsdk.ResultReason.SynthesizingAudioCompleted:
            return result.audio_duration.total_seconds()
        elif result.reason == speechsdk.ResultReason.Canceled:
            cancellation_details = result.cancellation_details
            if cancellation_details.reason == speechsdk.CancellationReason.Error:
                errmsg = f"Error details: {cancellation_details.error_details}"
                raise SynthesizeError(errmsg)
        raise SynthesizeError("Failed to get tts from azure")

    @classmethod
    def list_voices(cls) -> list[str]:
        return AZURE_SUPPORTED_VOICES

    @classmethod
    def get_command(cls) -> click.Command:
        @click.command()
        @click.option(
            "--speech-key",
            required=True,
            envvar="AZURE_SPEECH_KEY",
            show_envvar=True,
            help="The Azure Speech key.",
        )
        @click.option(
            "--speech-region",
            envvar="AZURE_SPEECH_REGION",
            show_envvar=True,
            required=True,
            help="The Azure Speech region.",
        )
        @click.option("--voice", default="en-US-AriaNeural", help="The voice to use.")
        @common_options(cls)
        def azure(
            speech_key: str, speech_region: str, voice: str, text: str, output: str
        ) -> None:
            speaker = cls(speech_key, speech_region, voice=voice)
            speaker.say(text, Path(output))

        return azure



import logging

import click

from .azure import AzureSpeaker
from .edge import EdgeSpeaker
from .google import GoogleSpeaker
from .openai import OpenAISpeaker
from .volc import VolcSpeaker


def setup_logger(debug: bool = False):
    logger = logging.getLogger("tetos")
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG if debug else logging.INFO)


@click.group(context_settings={"show_default": True})
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose logging.")
def tts(verbose: bool) -> None:
    setup_logger(verbose)


for speaker in (OpenAISpeaker, EdgeSpeaker, AzureSpeaker, VolcSpeaker, GoogleSpeaker):
    tts.add_command(speaker.get_command())


if __name__ == "__main__":
    tts()



import abc
from pathlib import Path
from typing import Any, Callable, TypeVar

import click

F = TypeVar("F", bound=Callable[..., Any])


class SynthesizeError(RuntimeError):
    pass


class Speaker(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def synthesize(self, text: str, out_file: Path) -> float:
        """Generate speech from text and save it to a file.

        Args:
            text (str): The text to synthesize.
            out_file (Path): The file to save the speech to.

        Returns:
            float: The duration of the speech in seconds.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def list_voices(cls) -> list[str]:
        """List the available voices for the speaker.

        Returns:
            list[str]: A list of voice names.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_command(cls) -> click.Command:
        """Return a Click command for the speaker.

        Returns:
            click.Command: The Click command.
        """
        raise NotImplementedError

    def say(self, text: str, out_file: Path | None = None) -> float:
        """A synchronous version of synthesize() that takes an optional
        playback argument to play the audio.
        """
        import anyio
        import click

        if out_file is None:
            out_file = Path("tts-output.mp3")

        result = anyio.run(self.synthesize, text, out_file)
        click.echo(f"Speech is generated successfully at {out_file}")
        return result


def common_options(cls: Speaker) -> Callable[[F], F]:
    def list_voices(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        click.echo("\n".join(cls.list_voices()))
        ctx.exit()

    def decorator(func: F) -> F:
        func = click.argument("text")(func)
        func = click.option(
            "--output",
            "-o",
            type=click.Path(dir_okay=False),
            default="tts-output.mp3",
            help="The output file.",
        )(func)
        func = click.option(
            "--list-voices",
            "-l",
            is_flag=True,
            help="List available voices.",
            is_eager=True,
            callback=list_voices,
            expose_value=False,
        )(func)
        return func

    return decorator
